stdin: Always set t->error correctly.
[paraslash.git] / stdin.c
diff --git a/stdin.c b/stdin.c
index bb922dd168f645ebf61b17c11726e705d2b8c6e8..abe69cafd4270ddf0e82ed9d2b30d3cf3603f43a 100644 (file)
--- a/stdin.c
+++ b/stdin.c
@@ -1,37 +1,44 @@
 /*
- * Copyright (C) 2006 Andre Noll <maan@systemlinux.org>
+ * Copyright (C) 2006-2009 Andre Noll <maan@systemlinux.org>
  *
  * Licensed under the GPL v2. For licencing details see COPYING.
  */
 
-/** \file stdin.c functions that deal with reading from stdin */
+/** \file stdin.c Functions that deal with reading from stdin. */
 
 #include <dirent.h> /* readdir() */
 #include <assert.h>
-
+#include <stdbool.h>
+#include <regex.h>
 
 #include "para.h"
-#include "string.h"
 #include "list.h"
 #include "sched.h"
 #include "fd.h"
 #include "error.h"
 #include "stdin.h"
+#include "buffer_tree.h"
+#include "string.h"
 
 /**
- * the pre_select function of the stdin task
+ * The pre_select function of the stdin task.
  *
- * \param s the scheduler this task was registered to
- * \param t the task structure of the stdin task
+ * \param s The scheduler this task was registered to.
+ * \param t The task structure of the stdin task.
  *
  * This function is always successful. If there is space left in the
  * buffer of the stdin task, it adds \p STDIN_FILENO to the read fd set
  * of \a s.
  */
-void stdin_pre_select(struct sched *s, struct task *t)
+static void stdin_pre_select(struct sched *s, struct task *t)
 {
-       struct stdin_task *sit = t->private_data;
-       t->ret = 1;
+       struct stdin_task *sit = container_of(t, struct stdin_task, task);
+
+       if (sit->output_error && *sit->output_error < 0) {
+               t->error = *sit->output_error;
+               return;
+       }
+       t->error = 0;
        sit->check_fd = 0;
        if (sit->loaded >= sit->bufsize)
                return;
@@ -39,43 +46,86 @@ void stdin_pre_select(struct sched *s, struct task *t)
        para_fd_set(STDIN_FILENO, &s->rfds, &s->max_fileno);
 }
 
-static void stdin_default_event_handler(struct task *t)
+static void stdin_pre_select_btr(struct sched *s, struct task *t)
 {
-       PARA_NOTICE_LOG("%p: %s\n", t, PARA_STRERROR(-t->ret));
-       unregister_task(t);
+       struct stdin_task *sit = container_of(t, struct stdin_task, task);
+       int ret;
+
+       t->error = 0;
+       ret = btr_node_status(sit->btrn, 0, BTR_NT_ROOT);
+       if (ret > 0)
+               para_fd_set(STDIN_FILENO, &s->rfds, &s->max_fileno);
+       else if (ret < 0) {
+               s->timeout.tv_sec = 0;
+               s->timeout.tv_usec = 1;
+       }
 }
 
 /**
- * the post select function of the stdin task
+ * The post select function of the stdin task.
  *
- * \param s the scheduler this task was registered to
- * \param t the task structure of the stdin task
+ * \param s The scheduler this task was registered to.
+ * \param t The task structure of the stdin task.
  *
  * This function checks if \p STDIN_FILENO was included by in the read fd set
  * of \a s during the previous pre_select call.  If yes, and \p STDIN_FILENO
- * appeears to be readable, data is read from stdin into the buffer of the
+ * appears to be readable, data is read from stdin into the buffer of the
  * stdin task.
  */
-void stdin_post_select(struct sched *s, struct task *t)
+static void stdin_post_select(struct sched *s, struct task *t)
 {
-       struct stdin_task *sit = t->private_data;
+       struct stdin_task *sit = container_of(t, struct stdin_task, task);
        ssize_t ret;
 
-       t->ret = 1;
+       if (sit->output_error && *sit->output_error < 0) {
+               t->error = *sit->output_error;
+               return;
+       }
+       t->error = 0;
        if (!sit->check_fd)
                return;
        if (!FD_ISSET(STDIN_FILENO, &s->rfds))
                return;
        ret = read(STDIN_FILENO, sit->buf + sit->loaded, sit->bufsize - sit->loaded);
        if (ret < 0)
-               t->ret = -E_STDIN_READ;
-       else if (ret > 0) {
+               t->error = ERRNO_TO_PARA_ERROR(errno);
+       else if (ret > 0)
                sit->loaded += ret;
-               t->ret = ret;
-       } else
-               t->ret = -E_STDIN_EOF;
-       if (t->ret < 0)
-               sit->eof = 1;
+       else
+               t->error = -E_STDIN_EOF;
+}
+
+#define STDIN_INPUT_BUFFER_SIZE (1024 * 32)
+static void stdin_post_select_btr(struct sched *s, struct task *t)
+{
+       struct stdin_task *sit = container_of(t, struct stdin_task, task);
+       ssize_t ret;
+       char *buf = NULL;
+
+       t->error = 0;
+       ret = btr_node_status(sit->btrn, 0, BTR_NT_ROOT);
+       if (ret < 0)
+               goto err;
+       if (ret == 0)
+               return;
+       if (!FD_ISSET(STDIN_FILENO, &s->rfds))
+               return;
+       buf = para_malloc(STDIN_INPUT_BUFFER_SIZE);
+       ret = read(STDIN_FILENO, buf, STDIN_INPUT_BUFFER_SIZE);
+       //PARA_CRIT_LOG("read ret: %d\n", ret);
+       if (ret <= 0) {
+               if (ret < 0)
+                       ret = -ERRNO_TO_PARA_ERROR(errno);
+               else
+                       ret = -E_STDIN_EOF;
+               goto err;
+       }
+       btr_add_output(buf, ret, sit->btrn);
+       return;
+err:
+       free(buf);
+       btr_remove_node(sit->btrn);
+       t->error = ret;
 }
 
 /**
@@ -83,20 +133,27 @@ void stdin_post_select(struct sched *s, struct task *t)
  *
  * \param sit The stdin task structure.
  *
* This fills in the pre/post select function poinzters of the task structure
* given by \a sit. It also sets up a default error handler which unregisters
* the task. Moreover, \a loaded and \a eof are set to zero and \a bufsize is
* initialized to 16 KB (but no buffer is allocated).
  */
 void stdin_set_defaults(struct stdin_task *sit)
 {
-       sit->bufsize = 16 * 1024,
-       sit->loaded = 0,
-       sit->eof = 0,
-       sit->task.pre_select = stdin_pre_select;
-       sit->task.post_select = stdin_post_select;
-       sit->task.event_handler = stdin_default_event_handler;
-       sit->task.private_data = sit;
-       mark_fd_nonblock(STDIN_FILENO);
+       int ret;
+
+       sit->bufsize = 32 * 1024;
+       if (sit->btrn) {
+               sit->task.pre_select = stdin_pre_select_btr;
+               sit->task.post_select = stdin_post_select_btr;
+       } else {
+               sit->task.pre_select = stdin_pre_select;
+               sit->task.post_select = stdin_post_select;
+       }
        sprintf(sit->task.status, "stdin reader");
+       ret = mark_fd_nonblocking(STDIN_FILENO);
+       if (ret >= 0)
+               return;
+       sit->output_error = NULL;
+       PARA_EMERG_LOG("%s\n", para_strerror(-ret));
+       exit(EXIT_FAILURE);
 }
@@ -83,20 +133,27 @@ void stdin_post_select(struct sched *s, struct task *t)
  * \param sit The stdin task structure.
  *
- * This fills in the pre/post select function poinzters of the task structure
- * given by \a sit. It also sets up a default error handler which unregisters
- * the task. Moreover, \a loaded and \a eof are set to zero and \a bufsize is
- * initialized to 16 KB (but no buffer is allocated).
+ * This fills in the pre/post select function pointers of the task structure
+ * given by \a sit. Moreover, the stdin file desctiptor is set to nonblocking
+ * mode and \a bufsize is initialized to 16 KB (but no buffer is allocated).
  */
 void stdin_set_defaults(struct stdin_task *sit)